# coding=utf-8

"""
The JCollectdCollector is capable of receiving Collectd network traffic
as sent by the JCollectd jvm agent (and child Collectd processes).

Reason for developing this collector is allowing to use JCollectd, without
the need for Collectd.

A few notes:

This collector starts a UDP server to receive data. This server runs in
a separate thread and puts it on a queue, waiting for the collect() method
to pull. Because of this setup, the collector interval parameter is of
less importance. What matters is the 'sendinterval' JCollectd parameter.

See https://github.com/emicklei/jcollectd for an up-to-date jcollect fork.

#### Dependencies

 * jcollectd sending metrics

"""


import threading
import re
import Queue

import diamond.collector
import diamond.metric

import collectd_network


ALIVE = True


class JCollectdCollector(diamond.collector.Collector):

    def __init__(self, *args, **kwargs):
        super(JCollectdCollector, self).__init__(*args, **kwargs)
        self.listener_thread = None

    def get_default_config(self):
        """
        Returns the default collector settings
        """
        config = super(JCollectdCollector, self).get_default_config()
        config.update({
            'path':     'jvm',
            'listener_host': '127.0.0.1',
            'listener_port': 25826,
        })
        return config

    def collect(self):
        if not self.listener_thread:
            self.start_listener()

        q = self.listener_thread.queue
        while True:
            try:
                dp = q.get(False)
                metric = self.make_metric(dp)
            except Queue.Empty:
                break
            self.publish_metric(metric)

    def start_listener(self):
        self.listener_thread = ListenerThread(self.config['listener_host'],
                                              self.config['listener_port'],
                                              self.log)
        self.listener_thread.start()

    def stop_listener(self):
        global ALIVE
        ALIVE = False
        self.listener_thread.join()
        self.log.error('Listener thread is shut down.')

    def make_metric(self, dp):

        path = ".".join((dp.host, self.config['path'], dp.name))

        if 'path_prefix' in self.config:
            prefix = self.config['path_prefix']
            if prefix:
                path = ".".join((prefix, path))

        if 'path_suffix' in self.config:
            suffix = self.config['path_suffix']
            if suffix:
                path = ".".join((path, suffix))

        if dp.is_counter:
            metric_type = "COUNTER"
        else:
            metric_type = "GAUGE"
        metric = diamond.metric.Metric(path, dp.value, dp.time,
                                       metric_type=metric_type)

        return metric

    def __del__(self):
        if self.listener_thread:
            self.stop_listener()


class ListenerThread(threading.Thread):

    def __init__(self, host, port, log, poll_interval=0.4):
        super(ListenerThread, self).__init__()
        self.name = 'JCollectdListener'  # thread name

        self.host = host
        self.port = port
        self.log = log
        self.poll_interval = poll_interval

        self.queue = Queue.Queue()

    def run(self):
        self.log.info('ListenerThread started on {0}:{1}(udp)'.format(
            self.host, self.port))

        rdr = collectd_network.Reader(self.host, self.port)

        try:
            while ALIVE:
                try:
                    items = rdr.interpret(poll_interval=self.poll_interval)
                    self.send_to_collector(items)
                except ValueError, e:
                    self.log.warn('Dropping bad packet: {0}'.format(e))
        except Exception, e:
            self.log.error('caught exception: type={0}, exc={1}'.format(type(e),
                                                                        e))

        self.log.info('ListenerThread - stop')

    def send_to_collector(self, items):
        if items is None:
            return

        for item in items:
            try:
                metric = self.transform(item)
                self.queue.put(metric)
            except Queue.Full:
                self.log.error('Queue to collector is FULL')
            except Exception, e:
                self.log.error('B00M! type={0}, exception={1}'.format(type(e),
                                                                      e))

    def transform(self, item):

        parts = []

        path = item.plugininstance
        # extract jvm name from 'logstash-MemoryPool Eden Space'
        if '-' in path:
            (jvm, tail) = path.split('-', 1)
            path = tail
        else:
            jvm = 'unnamed'

        # add JVM name
        parts.append(jvm)

        # add mbean name (e.g. 'java_lang')
        parts.append(item.plugin)

        # get typed mbean: 'MemoryPool Eden Space'
        if ' ' in path:
            (mb_type, mb_name) = path.split(' ', 1)
            parts.append(mb_type)
            parts.append(mb_name)
        else:
            parts.append(path)

        # add property name
        parts.append(item.typeinstance)

        # construct full path, from safe parts
        name = '.'.join([sanitize_word(part) for part in parts])

        if item[0][0] == 0:
            is_counter = True
        else:
            is_counter = False
        dp = Datapoint(item.host, item.time, name, item[0][1], is_counter)

        return dp


def sanitize_word(s):
    """Remove non-alphanumerical characters from metric word.
    And trim excessive underscores.
    """
    s = re.sub('[^\w-]+', '_', s)
    s = re.sub('__+', '_', s)
    return s.strip('_')


class Datapoint(object):

    def __init__(self, host, time, name, value, is_counter):
        self.host = host
        self.time = time
        self.name = name
        self.value = value
        self.is_counter = is_counter



# coding=utf-8

"""
Collect twemproxy (aka nutcracker) stats ( Modified from memcached collector )

#### Dependencies

 * json or simplejson

#### Example Configuration

TwemproxyCollector.conf

```
    enabled = True
    hosts = localhost:22222, app-1@localhost:22222, app-2@localhost:22222, etc
```

TO use a unix socket, set a host string like this

```
    hosts = /path/to/blah.sock, app-1@/path/to/bleh.sock,
```
"""

import diamond.collector
import socket
import re

try:
    import simplejson as json
except ImportError:
    import json


class TwemproxyCollector(diamond.collector.Collector):
    GAUGES = [
        'uptime',
        'curr_connections',
        'client_connections',
        'server_connections',
        'server_ejected_at',
        'in_queue',
        'in_queue_bytes',
        'out_queue',
        'out_queue_bytes'
    ]

    IGNORED = [
        'service',
        'source',
        'timestamp',
        'version'
    ]

    def get_default_config_help(self):
        config_help = super(TwemproxyCollector, self).get_default_config_help()
        config_help.update({
            'hosts': "List of hosts, and ports to collect. Set an alias by " +
            " prefixing the host:port with alias@",
        })
        return config_help

    def get_default_config(self):
        """
        Returns the default collector settings
        """
        config = super(TwemproxyCollector, self).get_default_config()
        config.update({
            'path':     'twemproxy',
            'hosts': ['localhost:22222']
        })
        return config

    def get_raw_stats(self, host, port):
        data = ''
        # connect
        try:
            if port is None:
                sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                sock.connect(host)
            else:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.connect((host, int(port)))

            stats_data = ''
            while True:
                data = sock.recv(1024)
                if not data:
                    break
                stats_data += data
            sock.close()

        except socket.error:
            self.log.exception('Failed to get stats from %s:%s',
                               host, port)

        try:
            return json.loads(stats_data)
        except (TypeError, ValueError):
            self.log.error("Unable to parse response from Twemproxy as a"
                           " json object")
            return False

    def get_stats(self, host, port):
        stats = {}
        pools = {}
        data = self.get_raw_stats(host, port)

        if data is None:
            self.log.error('Unable to import json')
            return {}

        stats = {}
        pools = {}
        for stat, value in data.iteritems():
            # Test if this is a pool
            if isinstance(value, dict):
                pool_name = stat.replace('.', '_')
                pools[pool_name] = {}
                for pool_stat, pool_value in value.iteritems():
                    # Test if this is a pool server
                    if isinstance(pool_value, dict):
                        server_name = pool_stat.replace('.', '_')
                        pools[pool_name][server_name] = {}
                        for server_stat, server_value in pool_value.iteritems():
                            pools[pool_name][server_name][server_stat] = \
                                int(server_value)
                    else:
                        pools[pool_name][pool_stat] = int(pool_value)
            else:
                if stat in self.IGNORED:
                    continue
                else:
                    stats[stat] = int(value)

        return stats, pools

    def collect(self):
        hosts = self.config.get('hosts')

        # Convert a string config value to be an array
        if isinstance(hosts, basestring):
            hosts = [hosts]

        for host in hosts:
            matches = re.search('((.+)\@)?([^:]+)(:(\d+))?', host)
            alias = matches.group(2)
            hostname = matches.group(3)
            port = matches.group(5)

            if alias is None:
                alias = hostname

            stats, pools = self.get_stats(hostname, port)

            for stat in stats:
                if stat in self.GAUGES:
                    self.publish_gauge(alias + "." + stat, stats[stat])
                else:
                    self.publish_counter(alias + "." + stat, stats[stat])

            # Pool stats
            for pool, pool_stats in pools.iteritems():
                for stat, stat_value in pool_stats.iteritems():
                    # Test if this is a pool server
                    if isinstance(stat_value, dict):
                        for server_stat, server_value in stat_value.iteritems():
                            if server_stat in self.GAUGES:
                                self.publish_gauge(
                                    alias + ".pools." + pool + ".servers." +
                                    stat + "." + server_stat, server_value)
                            else:
                                self.publish_counter(
                                    alias + ".pools." + pool + ".servers." +
                                    stat + "." + server_stat, server_value)
                    else:
                        if stat in self.GAUGES:
                            self.publish_gauge(
                                alias + ".pools." + pool + "." + stat,
                                stat_value)
                        else:
                            self.publish_counter(
                                alias + ".pools." + pool + "." + stat,
                                stat_value)



# coding=utf-8

"""
The CpuAcctCGroupCollector collects CPU Acct metric for cgroups

#### Dependencies

A mounted cgroup fs. Defaults to /sys/fs/cgroup/cpuacct/

"""

import diamond.collector
import os


class CpuAcctCgroupCollector(diamond.collector.Collector):

    def get_default_config_help(self):
        config_help = super(
            CpuAcctCgroupCollector, self).get_default_config_help()
        config_help.update({
            'path': """Directory path to where cpuacct is located,
defaults to /sys/fs/cgroup/cpuacct/. Redhat/CentOS/SL use /cgroup"""
        })
        return config_help

    def get_default_config(self):
        """
        Returns the default collector settings
        """
        config = super(CpuAcctCgroupCollector, self).get_default_config()
        config.update({
            'path':     '/sys/fs/cgroup/cpuacct/'
        })
        return config

    def collect(self):
        # find all cpuacct.stat files
        matches = []
        for root, dirnames, filenames in os.walk(self.config['path']):
            for filename in filenames:
                if filename == 'cpuacct.stat':
                    # matches will contain a tuple contain path to cpuacct.stat
                    # and the parent of the stat
                    parent = root.replace(self.config['path'],
                                          "").replace("/", ".")
                    if parent == '':
                        parent = 'system'
                    # If the parent starts with a dot, remove it
                    if parent[0] == '.':
                        parent = parent[1:]
                    matches.append((parent, os.path.join(root, filename)))

        # Read utime and stime from cpuacct files
        results = {}
        for match in matches:
            results[match[0]] = {}
            stat_file = open(match[1])
            elements = [line.split() for line in stat_file]
            for el in elements:
                results[match[0]][el[0]] = el[1]
                stat_file.close()

        # create metrics from collected utimes and stimes for cgroups
        for parent, cpuacct in results.iteritems():
            for key, value in cpuacct.iteritems():
                metric_name = '.'.join([parent, key])
                self.publish(metric_name, value, metric_type='GAUGE')
        return True



# coding=utf-8

import logging
import multiprocessing
import os
import signal
import sys
import time

try:
    from setproctitle import getproctitle, setproctitle
except ImportError:
    setproctitle = None

# Path Fix
sys.path.append(
    os.path.abspath(
        os.path.join(
            os.path.dirname(__file__), "../")))

from diamond.utils.classes import initialize_collector
from diamond.utils.classes import load_collectors
from diamond.utils.classes import load_dynamic_class
from diamond.utils.classes import load_handlers
from diamond.utils.classes import load_include_path

from diamond.utils.config import load_config

from diamond.utils.scheduler import collector_process
from diamond.utils.scheduler import handler_process

from diamond.handler.Handler import Handler

from diamond.utils.signals import signal_to_exception
from diamond.utils.signals import SIGHUPException


class Server(object):
    """
    Server class loads and starts Handlers and Collectors
    """

    def __init__(self, configfile):
        # Initialize Logging
        self.log = logging.getLogger('diamond')
        # Initialize Members
        self.configfile = configfile
        self.config = None
        self.handlers = []
        self.handler_queue = []
        self.modules = {}
        self.metric_queue = None

        # We do this weird process title swap around to get the sync manager
        # title correct for ps
        if setproctitle:
            oldproctitle = getproctitle()
            setproctitle('%s - SyncManager' % getproctitle())
        self.manager = multiprocessing.Manager()
        if setproctitle:
            setproctitle(oldproctitle)

    def run(self):
        """
        Load handler and collector classes and then start collectors
        """

        #######################################################################
        # Config
        #######################################################################
        self.config = load_config(self.configfile)

        collectors = load_collectors(self.config['server']['collectors_path'])
        metric_queue_size = int(self.config['server'].get('metric_queue_size',
                                                          16384))
        self.metric_queue = self.manager.Queue(maxsize=metric_queue_size)
        self.log.debug('metric_queue_size: %d', metric_queue_size)

        #######################################################################
        # Handlers
        #
        # TODO: Eventually move each handler to it's own process space?
        #######################################################################

        if 'handlers_path' in self.config['server']:
            handlers_path = self.config['server']['handlers_path']

            # Make an list if not one
            if isinstance(handlers_path, basestring):
                handlers_path = handlers_path.split(',')
                handlers_path = map(str.strip, handlers_path)
                self.config['server']['handlers_path'] = handlers_path

            load_include_path(handlers_path)

        if 'handlers' not in self.config['server']:
            self.log.critical('handlers missing from server section in config')
            sys.exit(1)

        handlers = self.config['server'].get('handlers')
        if isinstance(handlers, basestring):
            handlers = [handlers]

        # Prevent the Queue Handler from being a normal handler
        if 'diamond.handler.queue.QueueHandler' in handlers:
            handlers.remove('diamond.handler.queue.QueueHandler')

        self.handlers = load_handlers(self.config, handlers)

        QueueHandler = load_dynamic_class(
            'diamond.handler.queue.QueueHandler',
            Handler
        )

        self.handler_queue = QueueHandler(
            config=self.config, queue=self.metric_queue, log=self.log)

        process = multiprocessing.Process(
            name="Handlers",
            target=handler_process,
            args=(self.handlers, self.metric_queue, self.log),
        )

        process.daemon = True
        process.start()

        #######################################################################
        # Signals
        #######################################################################

        if hasattr(signal, 'SIGHUP'):
            signal.signal(signal.SIGHUP, signal_to_exception)

        #######################################################################

        while True:
            try:
                active_children = multiprocessing.active_children()
                running_processes = []
                for process in active_children:
                    running_processes.append(process.name)
                running_processes = set(running_processes)

                ##############################################################
                # Collectors
                ##############################################################

                running_collectors = []
                for collector, config in self.config['collectors'].iteritems():
                    if config.get('enabled', False) is not True:
                        continue
                    running_collectors.append(collector)
                running_collectors = set(running_collectors)

                # Collectors that are running but shouldn't be
                for process_name in running_processes - running_collectors:
                    if 'Collector' not in process_name:
                        continue
                    for process in active_children:
                        if process.name == process_name:
                            process.terminate()

                collector_classes = dict(
                    (cls.__name__.split('.')[-1], cls)
                    for cls in collectors.values()
                )

                load_delay = self.config['server'].get('collectors_load_delay',
                                                       1.0)
                for process_name in running_collectors - running_processes:
                    # To handle running multiple collectors concurrently, we
                    # split on white space and use the first word as the
                    # collector name to spin
                    collector_name = process_name.split()[0]

                    if 'Collector' not in collector_name:
                        continue

                    if collector_name not in collector_classes:
                        self.log.error('Can not find collector %s',
                                       collector_name)
                        continue

                    collector = initialize_collector(
                        collector_classes[collector_name],
                        name=process_name,
                        configfile=self.configfile,
                        handlers=[self.handler_queue])

                    if collector is None:
                        self.log.error('Failed to load collector %s',
                                       process_name)
                        continue

                    # Splay the loads
                    time.sleep(float(load_delay))

                    process = multiprocessing.Process(
                        name=process_name,
                        target=collector_process,
                        args=(collector, self.metric_queue, self.log)
                    )
                    process.daemon = True
                    process.start()

                ##############################################################

                time.sleep(1)

            except SIGHUPException:
                # ignore further SIGHUPs for now
                original_sighup_handler = signal.getsignal(signal.SIGHUP)
                signal.signal(signal.SIGHUP, signal.SIG_IGN)

                self.log.info('Reloading state due to HUP')
                self.config = load_config(self.configfile)
                collectors = load_collectors(
                    self.config['server']['collectors_path'])
                # restore SIGHUP handler
                signal.signal(signal.SIGHUP, original_sighup_handler)



"""
Collects metrics on queues and queue consumers from PgQ, a PostgreSQL-based
queueing mechanism (part of the Skytools utilities released by Skype.)

#### Dependencies

 * psycopg2

#### Example Configuration

```
enabled = True

[instances]

[[database1]]
dsn = postgresql://user:secret@localhost

[[database2]]
dsn = host=localhost port=5432 dbname=mydb
```
"""
try:
    import psycopg2
    import psycopg2.extras
except ImportError:
    psycopg2 = None

import diamond.collector


class PgQCollector(diamond.collector.Collector):

    def get_default_config_help(self):
        config_help = super(PgQCollector, self).get_default_config_help()
        config_help.update({
            "instances": "The databases to be monitored. Each should have a "
                         "`dsn` attribute, which must be a valid libpq "
                         "connection string."
        })
        return config_help

    def get_default_config(self):
        config = super(PgQCollector, self).get_default_config()
        config.update({
            'instances': {},
        })
        return config

    def collect(self):
        if psycopg2 is None:
            self.log.error('Unable to import module psycopg2')
            return None

        for instance, configuration in self.config['instances'].iteritems():
            connection = psycopg2.connect(configuration['dsn'])
            connection.set_isolation_level(
                psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT,
            )
            self._collect_for_instance(instance, connection)

    def _collect_for_instance(self, instance, connection):
        "Collects metrics for a named connection."
        with connection.cursor() as cursor:
            for queue, metrics in self.get_queue_info(instance, cursor):
                for name, metric in metrics.items():
                    self.publish('.'.join((instance, queue, name)), metric)

        with connection.cursor() as cursor:
            consumers = self.get_consumer_info(instance, cursor)
            for queue, consumer, metrics in consumers:
                for name, metric in metrics.items():
                    key_parts = (instance, queue, 'consumers', consumer, name)
                    self.publish('.'.join(key_parts), metric)

    QUEUE_INFO_STATEMENT = """
        SELECT
            queue_name,
            EXTRACT(epoch from ticker_lag),
            ev_per_sec
        FROM pgq.get_queue_info()
    """

    def get_queue_info(self, instance, cursor):
        "Collects metrics for all queues on the connected database."
        cursor.execute(self.QUEUE_INFO_STATEMENT)
        for queue_name, ticker_lag, ev_per_sec in cursor:
            yield queue_name, {
                'ticker_lag': ticker_lag,
                'ev_per_sec': ev_per_sec,
            }

    CONSUMER_INFO_STATEMENT = """
        SELECT
            queue_name,
            consumer_name,
            EXTRACT(epoch from lag),
            pending_events,
            EXTRACT(epoch from last_seen)
        FROM pgq.get_consumer_info()
    """

    def get_consumer_info(self, instance, cursor):
        "Collects metrics for all consumers on the connected database."
        cursor.execute(self.CONSUMER_INFO_STATEMENT)
        for queue_name, consumer_name, lag, pending_events, last_seen in cursor:
            yield queue_name, consumer_name, {
                'lag': lag,
                'pending_events': pending_events,
                'last_seen': last_seen,
            }



# coding=utf-8

"""
Collect stats from Apache HTTPD server using mod_status

#### Dependencies

 * mod_status
 * httplib
 * urlparse

"""

import re
import httplib
import urlparse
import diamond.collector


class HttpdCollector(diamond.collector.Collector):

    def process_config(self):
        super(HttpdCollector, self).process_config()
        if 'url' in self.config:
            self.config['urls'].append(self.config['url'])

        self.urls = {}
        if isinstance(self.config['urls'], basestring):
            self.config['urls'] = self.config['urls'].split(',')

        for url in self.config['urls']:
            # Handle the case where there is a trailing comman on the urls list
            if len(url) == 0:
                continue
            if ' ' in url:
                parts = url.split(' ')
                self.urls[parts[0]] = parts[1]
            else:
                self.urls[''] = url

    def get_default_config_help(self):
        config_help = super(HttpdCollector, self).get_default_config_help()
        config_help.update({
            'urls': "Urls to server-status in auto format, comma seperated," +
                    " Format 'nickname http://host:port/server-status?auto, " +
                    ", nickname http://host:port/server-status?auto, etc'",
        })
        return config_help

    def get_default_config(self):
        """
        Returns the default collector settings
        """
        config = super(HttpdCollector, self).get_default_config()
        config.update({
            'path':     'httpd',
            'urls':     ['localhost http://localhost:8080/server-status?auto']
        })
        return config

    def collect(self):
        for nickname in self.urls.keys():
            url = self.urls[nickname]

            try:
                while True:

                    # Parse Url
                    parts = urlparse.urlparse(url)

                    # Parse host and port
                    endpoint = parts[1].split(':')
                    if len(endpoint) > 1:
                        service_host = endpoint[0]
                        service_port = int(endpoint[1])
                    else:
                        service_host = endpoint[0]
                        service_port = 80

                    # Setup Connection
                    connection = httplib.HTTPConnection(service_host,
                                                        service_port)

                    url = "%s?%s" % (parts[2], parts[4])

                    connection.request("GET", url)
                    response = connection.getresponse()
                    data = response.read()
                    headers = dict(response.getheaders())
                    if (('location' not in headers or
                         headers['location'] == url)):
                        connection.close()
                        break
                    url = headers['location']
                    connection.close()
            except Exception, e:
                self.log.error(
                    "Error retrieving HTTPD stats for host %s:%s, url '%s': %s",
                    service_host, str(service_port), url, e)
                continue

            exp = re.compile('^([A-Za-z ]+):\s+(.+)$')
            for line in data.split('\n'):
                if line:
                    m = exp.match(line)
                    if m:
                        k = m.group(1)
                        v = m.group(2)

                        # IdleWorkers gets determined from the scoreboard
                        if k == 'IdleWorkers':
                            continue

                        if k == 'Scoreboard':
                            for sb_kv in self._parseScoreboard(v):
                                self._publish(nickname, sb_kv[0], sb_kv[1])
                        else:
                            self._publish(nickname, k, v)

    def _publish(self, nickname, key, value):

        metrics = ['ReqPerSec', 'BytesPerSec', 'BytesPerReq', 'BusyWorkers',
                   'Total Accesses', 'IdleWorkers', 'StartingWorkers',
                   'ReadingWorkers', 'WritingWorkers', 'KeepaliveWorkers',
                   'DnsWorkers', 'ClosingWorkers', 'LoggingWorkers',
                   'FinishingWorkers', 'CleanupWorkers']

        metrics_precision = ['ReqPerSec', 'BytesPerSec', 'BytesPerReq']

        if key in metrics:
            # Get Metric Name
            presicion_metric = False
            metric_name = "%s" % re.sub('\s+', '', key)
            if metric_name in metrics_precision:
                presicion_metric = 1

            # Prefix with the nickname?
            if len(nickname) > 0:
                metric_name = nickname + '.' + metric_name

            # Use precision for ReqPerSec BytesPerSec BytesPerReq
            if presicion_metric:
                # Get Metric Value
                metric_value = "%f" % float(value)

                # Publish Metric
                self.publish(metric_name, metric_value, precision=5)
            else:
                # Get Metric Value
                metric_value = "%d" % float(value)

                # Publish Metric
                self.publish(metric_name, metric_value)

    def _parseScoreboard(self, sb):

        ret = []

        ret.append(('IdleWorkers', sb.count('_')))
        ret.append(('ReadingWorkers', sb.count('R')))
        ret.append(('WritingWorkers', sb.count('W')))
        ret.append(('KeepaliveWorkers', sb.count('K')))
        ret.append(('DnsWorkers', sb.count('D')))
        ret.append(('ClosingWorkers', sb.count('C')))
        ret.append(('LoggingWorkers', sb.count('L')))
        ret.append(('FinishingWorkers', sb.count('G')))
        ret.append(('CleanupWorkers', sb.count('I')))

        return ret



""" This collector retrieves disk information from netapp filers

    !Thread-safe!

    The following metrics are measured in this diamond collector:
        * Number of disk in spare pool and not zeroed.
        * Number of spare disks per disk type, ie SAS/SATA
        * Number of disk in 'maintenance center'
        * monitor CP
        * Monitor average disk busyness per aggregate.

        Config File:
            [devices]
            [[filer-corp-201]] <--- You can have as many filers as you want!
            ip = 192.168.1.45
            user = netapp_monitor
            password = b0bl0rd!

"""

import diamond.collector
import time
from diamond.metric import Metric

try:
    import xml.etree.ElementTree as ET
except ImportError:
    import cElementTree as ET

try:
    from netappsdk.NaServer import *
    from netappsdk.NaElement import *
    netappsdk = 0
except ImportError:
    netappsdk = 1

__author__ = 'peter@phyn3t.com'


class netappDiskCol(object):
    """ Our netappDisk Collector
    """

    def __init__(self, device, ip, user, password, parent):
        """ Collectors our metrics for our netapp filer
        """

        self.device = device
        self.ip = ip
        self.netapp_user = user
        self.netapp_password = password
        self.path_prefix = parent[0]
        self.publish_metric = parent[1]
        self.log = parent[2]
        self._netapp_login()

        # Grab our netapp XML
        disk_xml = self.get_netapp_elem(
            NaElement('disk-list-info'), 'disk-details')
        storage_disk_xml = self.get_netapp_elem(
            NaElement('storage-disk-get-iter'), 'attributes-list')

        # Our metric collection and publishing goes here
        self.zero_disk(disk_xml)
        self.spare_disk(disk_xml)
        self.maintenance_center(storage_disk_xml)
        self.consistency_point()
        self.agr_busy()

    def agr_busy(self):
        """ Collector for average disk busyness per aggregate

            As of Nov 22nd 2013 there is no API call for agr busyness.
            You have to collect all disk busyness and then compute agr
            busyness. #fml

        """

        c1 = {}  # Counters from time a
        c2 = {}  # Counters from time b
        disk_results = {}  # Disk busyness results %
        agr_results = {}  # Aggregate busyness results $
        names = ['disk_busy', 'base_for_disk_busy', 'raid_name',
                 'base_for_disk_busy', 'instance_uuid']
        netapp_api = NaElement('perf-object-get-instances')
        netapp_api.child_add_string('objectname', 'disk')
        disk_1 = self.get_netapp_elem(netapp_api, 'instances')
        time.sleep(1)
        disk_2 = self.get_netapp_elem(netapp_api, 'instances')

        for instance_data in disk_1:
            temp = {}
            for element in instance_data.findall(".//counters/counter-data"):
                if element.find('name').text in names:
                    temp[element.find('name').text] = element.find(
                        'value').text

            agr_name = temp['raid_name']
            agr_name = agr_name[agr_name.find('/', 0):agr_name.find('/', 1)]
            temp['raid_name'] = agr_name.lstrip('/')
            c1[temp.pop('instance_uuid')] = temp

        for instance_data in disk_2:
            temp = {}
            for element in instance_data.findall(".//counters/counter-data"):
                if element.find('name').text in names:
                    temp[element.find('name').text] = element.find(
                        'value').text

            agr_name = temp['raid_name']
            agr_name = agr_name[agr_name.find('/', 0):agr_name.find('/', 1)]
            temp['raid_name'] = agr_name.lstrip('/')
            c2[temp.pop('instance_uuid')] = temp

        for item in c1:
            t_c1 = int(c1[item]['disk_busy'])  # time_counter_1
            t_b1 = int(c1[item]['base_for_disk_busy'])  # time_base_1
            t_c2 = int(c2[item]['disk_busy'])
            t_b2 = int(c2[item]['base_for_disk_busy'])

            disk_busy = 100 * (t_c2 - t_c1) / (t_b2 - t_b1)

            if c1[item]['raid_name'] in disk_results:
                disk_results[c1[item]['raid_name']].append(disk_busy)
            else:
                disk_results[c1[item]['raid_name']] = [disk_busy]

        for aggregate in disk_results:
            agr_results[aggregate] = \
                sum(disk_results[aggregate]) / len(disk_results[aggregate])

        for aggregate in agr_results:
            self.push('avg_busy', 'aggregate.' + aggregate,
                      agr_results[aggregate])

    def consistency_point(self):
        """ Collector for getting count of consistancy points
        """

        cp_delta = {}
        xml_path = 'instances/instance-data/counters'
        netapp_api = NaElement('perf-object-get-instances')
        netapp_api.child_add_string('objectname', 'wafl')
        instance = NaElement('instances')
        instance.child_add_string('instance', 'wafl')
        counter = NaElement('counters')
        counter.child_add_string('counter', 'cp_count')
        netapp_api.child_add(counter)
        netapp_api.child_add(instance)

        cp_1 = self.get_netapp_elem(netapp_api, xml_path)
        time.sleep(3)
        cp_2 = self.get_netapp_elem(netapp_api, xml_path)

        for element in cp_1:
            if element.find('name').text == 'cp_count':
                cp_1 = element.find('value').text.rsplit(',')
                break
        for element in cp_2:
            if element.find('name').text == 'cp_count':
                cp_2 = element.find('value').text.rsplit(',')
                break

        if not type(cp_2) is list or not type(cp_1) is list:
            log.error("consistency point data not available for filer: %s"
                      % self.device)
            return

        cp_1 = {
            'wafl_timer': cp_1[0],
            'snapshot': cp_1[1],
            'wafl_avail_bufs': cp_1[2],
            'dirty_blk_cnt': cp_1[3],
            'full_nv_log': cp_1[4],
            'b2b': cp_1[5],
            'flush_gen': cp_1[6],
            'sync_gen': cp_1[7],
            'def_b2b': cp_1[8],
            'con_ind_pin': cp_1[9],
            'low_mbuf_gen': cp_1[10],
            'low_datavec_gen': cp_1[11]
        }

        cp_2 = {
            'wafl_timer': cp_2[0],
            'snapshot': cp_2[1],
            'wafl_avail_bufs': cp_2[2],
            'dirty_blk_cnt': cp_2[3],
            'full_nv_log': cp_2[4],
            'b2b': cp_2[5],
            'flush_gen': cp_2[6],
            'sync_gen': cp_2[7],
            'def_b2b': cp_2[8],
            'con_ind_pin': cp_2[9],
            'low_mbuf_gen': cp_2[10],
            'low_datavec_gen': cp_2[11]
        }

        for item in cp_1:
            c1 = int(cp_1[item])
            c2 = int(cp_2[item])
            cp_delta[item] = c2 - c1

        for item in cp_delta:
            self.push(item + '_CP', 'system.system', cp_delta[item])

    def maintenance_center(self, storage_disk_xml=None):
        """ Collector for how many disk(s) are in NetApp maintenance center

            For more information on maintenance center please see:
              bit.ly/19G4ptr

        """

        disk_in_maintenance = 0

        for filer_disk in storage_disk_xml:
            disk_status = filer_disk.find('disk-raid-info/container-type')
            if disk_status.text == 'maintenance':
                disk_in_maintenance += 1

        self.push('maintenance_disk', 'disk', disk_in_maintenance)

    def zero_disk(self, disk_xml=None):
        """ Collector and publish not zeroed disk metrics
        """

        troubled_disks = 0
        for filer_disk in disk_xml:
            raid_state = filer_disk.find('raid-state').text
            if not raid_state == 'spare':
                continue
            is_zeroed = filer_disk.find('is-zeroed').text

            if is_zeroed == 'false':
                troubled_disks += 1
        self.push('not_zeroed', 'disk', troubled_disks)

    def spare_disk(self, disk_xml=None):
        """ Number of spare disk per type.

            For example: storage.ontap.filer201.disk.SATA

        """

        spare_disk = {}
        disk_types = set()

        for filer_disk in disk_xml:
            disk_types.add(filer_disk.find('effective-disk-type').text)
            if not filer_disk.find('raid-state').text == 'spare':
                continue

            disk_type = filer_disk.find('effective-disk-type').text
            if disk_type in spare_disk:
                spare_disk[disk_type] += 1
            else:
                spare_disk[disk_type] = 1

        for disk_type in disk_types:
            if disk_type in spare_disk:
                self.push('spare_' + disk_type, 'disk', spare_disk[disk_type])
            else:
                self.push('spare_' + disk_type, 'disk', 0)

    def get_netapp_elem(self, netapp_api=None, sub_element=None):
        """ Retrieve netapp elem
        """

        netapp_data = self.server.invoke_elem(netapp_api)

        if netapp_data.results_status() == 'failed':
            self.log.error(
                'While using netapp API failed to retrieve '
                'disk-list-info for netapp filer %s' % self.device)
            print netapp_data.sprintf()
            return
        netapp_xml = \
            ET.fromstring(netapp_data.sprintf()).find(sub_element)

        return netapp_xml

    def _netapp_login(self):
        """ Login to our netapp filer
        """

        self.server = NaServer(self.ip, 1, 3)
        self.server.set_transport_type('HTTPS')
        self.server.set_style('LOGIN')
        self.server.set_admin_user(self.netapp_user, self.netapp_password)

    def push(self, metric_name=None, type=None, metric_value=None):
        """ Ship that shit off to graphite broski
        """

        graphite_path = self.path_prefix
        graphite_path += '.' + self.device + '.' + type
        graphite_path += '.' + metric_name

        metric = Metric(
            graphite_path,
            metric_value,
            precision=4,
            host=self.device)

        self.publish_metric(metric)


class netappDisk(diamond.collector.Collector):
    """ Netapp disk diamond scheduler
    """

    running = set()

    def collect(self, device, ip, user, password):
        """ Collectors our metrics for our netapp filer --START HERE--
        """

        if netappsdk:
            self.log.error(
                'Failed to import netappsdk.NaServer or netappsdk.NaElement')
            return

        if device in self.running:
            return

        self.running.add(device)
        parent = (self.config['path_prefix'], self.publish_metric, self.log)

        netappDiskCol(device, ip, user, password, parent)
        self.running.remove(device)



# coding=utf-8

"""
The NfsCollector collects nfs utilization metrics using /proc/net/rpc/nfs.

#### Dependencies

 * /proc/net/rpc/nfs

"""

import diamond.collector
import os


class NfsCollector(diamond.collector.Collector):

    PROC = '/proc/net/rpc/nfs'

    def get_default_config_help(self):
        config_help = super(NfsCollector, self).get_default_config_help()
        config_help.update({
        })
        return config_help

    def get_default_config(self):
        """
        Returns the default collector settings
        """
        config = super(NfsCollector, self).get_default_config()
        config.update({
            'path':     'nfs'
        })
        return config

    def collect(self):
        """
        Collect stats
        """
        if os.access(self.PROC, os.R_OK):

            results = {}
            # Open file
            file = open(self.PROC)

            for line in file:
                line = line.split()

                if line[0] == 'net':
                    results['net.packets'] = line[1]
                    results['net.udpcnt'] = line[2]
                    results['net.tcpcnt'] = line[3]
                    results['net.tcpconn'] = line[4]
                elif line[0] == 'rpc':
                    results['rpc.calls'] = line[1]
                    results['rpc.retrans'] = line[2]
                    results['rpc.authrefrsh'] = line[3]
                elif line[0] == 'proc2':
                    line.pop(1)     # remove column-cnt field
                    results['v2.null'] = line[1]
                    results['v2.getattr'] = line[2]
                    results['v2.setattr'] = line[3]
                    results['v2.root'] = line[4]
                    results['v2.lookup'] = line[5]
                    results['v2.readlink'] = line[6]
                    results['v2.read'] = line[7]
                    results['v2.wrcache'] = line[8]
                    results['v2.write'] = line[9]
                    results['v2.create'] = line[10]
                    results['v2.remove'] = line[11]
                    results['v2.rename'] = line[12]
                    results['v2.link'] = line[13]
                    results['v2.symlink'] = line[14]
                    results['v2.mkdir'] = line[15]
                    results['v2.rmdir'] = line[16]
                    results['v2.readdir'] = line[17]
                    results['v2.fsstat'] = line[18]
                elif line[0] == 'proc3':
                    line.pop(1)     # remove column-cnt field
                    results['v3.null'] = line[1]
                    results['v3.getattr'] = line[2]
                    results['v3.setattr'] = line[3]
                    results['v3.lookup'] = line[4]
                    results['v3.access'] = line[5]
                    results['v3.readlink'] = line[6]
                    results['v3.read'] = line[7]
                    results['v3.write'] = line[8]
                    results['v3.create'] = line[9]
                    results['v3.mkdir'] = line[10]
                    results['v3.symlink'] = line[11]
                    results['v3.mknod'] = line[12]
                    results['v3.remove'] = line[13]
                    results['v3.rmdir'] = line[14]
                    results['v3.rename'] = line[15]
                    results['v3.link'] = line[16]
                    results['v3.readdir'] = line[17]
                    results['v3.readdirplus'] = line[18]
                    results['v3.fsstat'] = line[19]
                    results['v3.fsinfo'] = line[20]
                    results['v3.pathconf'] = line[21]
                    results['v3.commit'] = line[22]
                elif line[0] == 'proc4':
                    line.pop(1)     # remove column-cnt field
                    results['v4.null'] = line[1]
                    results['v4.read'] = line[2]
                    results['v4.write'] = line[3]
                    results['v4.commit'] = line[4]
                    results['v4.open'] = line[5]
                    results['v4.open_conf'] = line[6]
                    results['v4.open_noat'] = line[7]
                    results['v4.open_dgrd'] = line[8]
                    results['v4.close'] = line[9]
                    results['v4.setattr'] = line[10]
                    results['v4.fsinfo'] = line[11]
                    results['v4.renew'] = line[12]
                    results['v4.setclntid'] = line[13]
                    results['v4.confirm'] = line[14]
                    results['v4.lock'] = line[15]
                    results['v4.lockt'] = line[16]
                    results['v4.locku'] = line[17]
                    results['v4.access'] = line[18]
                    results['v4.getattr'] = line[19]
                    results['v4.lookup'] = line[20]
                    results['v4.lookup_root'] = line[21]
                    results['v4.remove'] = line[22]
                    results['v4.rename'] = line[23]
                    results['v4.link'] = line[24]
                    results['v4.symlink'] = line[25]
                    results['v4.create'] = line[26]
                    results['v4.pathconf'] = line[27]
                    results['v4.statfs'] = line[28]
                    results['v4.readlink'] = line[29]
                    results['v4.readdir'] = line[30]
                    try:
                        results['v4.server_caps'] = line[31]
                    except IndexError:
                        pass
                    try:
                        results['v4.delegreturn'] = line[32]
                    except IndexError:
                        pass
                    try:
                        results['v4.getacl'] = line[33]
                    except IndexError:
                        pass
                    try:
                        results['v4.setacl'] = line[34]
                    except IndexError:
                        pass
                    try:
                        results['v4.fs_locations'] = line[35]
                    except IndexError:
                        pass
                    try:
                        results['v4.rel_lkowner'] = line[36]
                    except IndexError:
                        pass
                    try:
                        results['v4.exchange_id'] = line[37]
                    except IndexError:
                        pass
                    try:
                        results['v4.create_ses'] = line[38]
                    except IndexError:
                        pass
                    try:
                        results['v4.destroy_ses'] = line[39]
                    except IndexError:
                        pass
                    try:
                        results['v4.sequence'] = line[40]
                    except IndexError:
                        pass
                    try:
                        results['v4.get_lease_t'] = line[41]
                    except IndexError:
                        pass
                    try:
                        results['v4.reclaim_comp'] = line[42]
                    except IndexError:
                        pass
                    try:
                        results['v4.layoutget'] = line[43]
                    except IndexError:
                        pass
                    try:
                        results['v4.layoutcommit'] = line[44]
                    except IndexError:
                        pass
                    try:
                        results['v4.layoutreturn'] = line[45]
                    except IndexError:
                        pass
                    try:
                        results['v4.getdevlist'] = line[46]
                    except IndexError:
                        pass
                    try:
                        results['v4.getdevinfo'] = line[47]
                    except IndexError:
                        pass
                    try:
                        results['v4.ds_write'] = line[48]
                    except IndexError:
                        pass
                    try:
                        results['v4.ds_commit'] = line[49]
                    except IndexError:
                        pass
                    try:
                        results['v4.getdevlist'] = line[50]
                    except IndexError:
                        pass

            # Close File
            file.close()

            for stat in results.keys():
                metric_name = stat
                metric_value = long(float(results[stat]))
                metric_value = self.derivative(metric_name, metric_value)
                self.publish(metric_name, metric_value, precision=3)
            return True

        return False



# coding=utf-8

"""
Send metrics to signalfx

#### Dependencies

 * urllib2


#### Configuration
Enable this handler

 * handers = diamond.handler.httpHandler.SignalfxHandler

 * auth_token = SIGNALFX_AUTH_TOKEN
 * batch_size = [optional | 300 ] will wait for this many requests before
     posting
"""

from Handler import Handler
from diamond.util import get_diamond_version
import json
import logging
import time
import urllib2


class SignalfxHandler(Handler):

    # Inititalize Handler with url and batch size
    def __init__(self, config=None):
        Handler.__init__(self, config)
        self.metrics = []
        self.batch_size = int(self.config['batch'])
        self.url = self.config['url']
        self.auth_token = self.config['auth_token']
        self.batch_max_interval = self.config['batch_max_interval']
        self.resetBatchTimeout()
        if self.auth_token == "":
            logging.error("Failed to load Signalfx module")
            return

    def resetBatchTimeout(self):
        self.batch_max_timestamp = int(time.time() + self.batch_max_interval)

    def get_default_config_help(self):
        """
        Returns the help text for the configuration options for this handler
        """
        config = super(SignalfxHandler, self).get_default_config_help()

        config.update({
            'url': 'Where to send metrics',
            'batch': 'How many to store before sending',
            'auth_token': 'Org API token to use when sending metrics',
        })

        return config

    def get_default_config(self):
        """
        Return the default config for the handler
        """
        config = super(SignalfxHandler, self).get_default_config()

        config.update({
            'url': 'https://ingest.signalfx.com/v2/datapoint',
            'batch': 300,
            # Don't wait more than 10 sec between pushes
            'batch_max_interval': 10,
            'auth_token': '',
        })

        return config

    def process(self, metric):
        """
        Queue a metric.  Flushing queue if batch size reached
        """
        self.metrics.append(metric)
        if self.should_flush():
            self._send()

    def should_flush(self):
        return len(self.metrics) >= self.batch_size or \
            time.time() >= self.batch_max_timestamp

    def into_signalfx_point(self, metric):
        """
        Convert diamond metric into something signalfx can understand
        """
        dims = {
            "collector": metric.getCollectorPath(),
            "prefix": metric.getPathPrefix(),
        }
        if metric.host is not None and metric.host != "":
            dims["host"] = metric.host

        return {
            "metric": metric.getMetricPath(),
            "value": metric.value,
            "dimensions": dims,
            # We expect ms timestamps
            "timestamp": metric.timestamp * 1000,
        }

    def flush(self):
        """Flush metrics in queue"""
        self._send()

    def user_agent(self):
        """
        HTTP user agent
        """
        return "Diamond: %s" % get_diamond_version()

    def _send(self):
        # Potentially use protobufs in the future
        postDictionary = {}
        for metric in self.metrics:
            t = metric.metric_type.lower()
            if t not in postDictionary:
                postDictionary[t] = []
            postDictionary[t].append(self.into_signalfx_point(metric))

        self.metrics = []
        postBody = json.dumps(postDictionary)
        logging.debug("Body is %s", postBody)
        req = urllib2.Request(self.url, postBody,
                              {"Content-type": "application/json",
                               "X-SF-TOKEN": self.auth_token,
                               "User-Agent": self.user_agent()})
        self.resetBatchTimeout()
        try:
            urllib2.urlopen(req)
        except urllib2.URLError:
            logging.exception("Unable to post signalfx metrics")
            return



# coding=utf-8

"""
Collects Cassandra JMX metrics from the Jolokia Agent.  Extends the
JolokiaCollector to interpret Histogram beans with information about the
distribution of request latencies.

#### Example Configuration
CassandraJolokiaCollector uses a regular expression to determine which
attributes represent histograms. This regex can be overridden by providing a
`histogram_regex` in your configuration.  You can also override `percentiles` to
collect specific percentiles from the histogram statistics.  The format is shown
below with the default values.

CassandraJolokiaCollector.conf

```
    percentiles '50,95,99'
    histogram_regex '.*HistogramMicros$'
```
"""

from jolokia import JolokiaCollector
import math
import string
import re


class CassandraJolokiaCollector(JolokiaCollector):
    # override to allow setting which percentiles will be collected

    def get_default_config_help(self):
        config_help = super(CassandraJolokiaCollector,
                            self).get_default_config_help()
        config_help.update({
            'percentiles':
            'Comma separated list of percentiles to be collected '
            '(e.g., "50,95,99").',
            'histogram_regex':
            'Filter to only process attributes that match this regex'
        })
        return config_help

    # override to allow setting which percentiles will be collected
    def get_default_config(self):
        config = super(CassandraJolokiaCollector, self).get_default_config()
        config.update({
            'percentiles': ['50', '95', '99'],
            'histogram_regex': '.*HistogramMicros$'
        })
        return config

    def __init__(self, *args, **kwargs):
        super(CassandraJolokiaCollector, self).__init__(*args, **kwargs)
        self.offsets = self.create_offsets(91)
        self.update_config(self.config)

    def update_config(self, config):
        if 'percentiles' in config:
            self.percentiles = map(int, config['percentiles'])
        if 'histogram_regex' in config:
            self.histogram_regex = re.compile(config['histogram_regex'])

    # override: Interpret beans that match the `histogram_regex` as histograms,
    # and collect percentiles from them.
    def interpret_bean_with_list(self, prefix, values):
        if not self.histogram_regex.match(prefix):
            return

        buckets = values
        offsets = self.offsets
        for percentile in self.percentiles:
            value = self.compute_percentile(offsets, buckets, percentile)
            cleaned_key = self.clean_up("%s.p%s" % (prefix, percentile))
            self.publish(cleaned_key, value)

    # Adapted from Cassandra docs:
    # https://bit.ly/13M5JPE
    # The index corresponds to the x-axis in a histogram.  It represents buckets
    # of values, which are a series of ranges. Each offset includes the range of
    # values greater than the previous offset and less than or equal to the
    # current offset. The offsets start at 1 and each subsequent offset is
    # calculated by multiplying the previous offset by 1.2, rounding up, and
    # removing duplicates. The offsets can range from 1 to approximately 25
    # million, with less precision as the offsets get larger.
    def compute_percentile(self, offsets, buckets, percentile_int):
        non_zero_points_sum = sum(buckets)
        if non_zero_points_sum is 0:
            return 0
        middle_point_index = math.floor(
            non_zero_points_sum * (percentile_int / float(100)))

        points_seen = 0
        for index, bucket in enumerate(buckets):
            points_seen += bucket
            if points_seen >= middle_point_index:
                return round((offsets[index] - offsets[index - 1]) / 2)

    # Returns a list of offsets for `n` buckets.
    def create_offsets(self, bucket_count):
        last_num = 1
        offsets = [last_num]

        for index in range(bucket_count):
            next_num = round(last_num * 1.2)
            if next_num == last_num:
                next_num += 1
            offsets.append(next_num)
            last_num = next_num

        return offsets



# coding=utf-8

"""
The FlutendCollector monitors fluentd and data about the kinesis stream.

#### Dependencies

 * flutend

#### Example config

```
    enabled = True
    host = localhost
    port = 24220
    [[[collect]]]
    kinesis = buffer_queue_length, buffer_total_queued_size, retry_count
```

"""

import diamond.collector
import urllib2
import json


class FluentdCollector(diamond.collector.Collector):

    API_PATH = '/api/plugins.json'

    def get_default_config_help(self):
        config_help = super(FluentdCollector, self).get_default_config_help()
        config_help.update({
            'host': 'Fluentd host',
            'port': 'Fluentd port',
            'collect': 'Plugins and their metrics to collect'
        })
        return config_help

    def get_default_config(self):
        config = super(FluentdCollector, self).get_default_config()
        config.update({
            'host': 'localhost',
            'port': '24220',
            'path': 'fluentd',
            'collect': {}
        })
        return config

    def collect(self):
        params = (self.config['host'], self.config['port'], self.API_PATH)
        url = "http://%s:%s/%s" % params

        res = urllib2.urlopen(url)
        data = json.load(res)

        result = self.parse_api_output(data)
        for r in result:
            self.publish(r[0], r[1])

    def parse_api_output(self, status):
        result = []
        for p in status.get('plugins'):
            if p['type'] in self.config['collect'].keys():
                for m in self.config['collect'].get(p['type']):
                    tag = ".".join([p['type'], m])
                    result.append((tag, p.get(m)))
        return result



# coding=utf-8

"""
[Librato](http://librato.com) is an infrastructure software as a service company
dedicated to delivering beautiful, easy to use tools that make managing your
operations more fun and efficient.

#### Dependencies

 * [librato-metrics](https://github.com/librato/python-librato)


"""

from Handler import Handler
import logging
import time
import re

try:
    import librato
except ImportError:
    librato = None


class LibratoHandler(Handler):

    def __init__(self, config=None):
        """
        Create a new instance of the LibratoHandler class
        """
        # Initialize Handler
        Handler.__init__(self, config)
        logging.debug("Initialized Librato handler.")

        if librato is None:
            logging.error("Failed to load librato module")
            return

        # Initialize Options
        api = librato.connect(self.config['user'],
                              self.config['apikey'])
        self.queue = api.new_queue()
        self.queue_max_size = int(self.config['queue_max_size'])
        self.queue_max_interval = int(self.config['queue_max_interval'])
        self.queue_max_timestamp = int(time.time() + self.queue_max_interval)
        self.current_n_measurements = 0

        # If a user leaves off the ending comma, cast to a array for them
        include_filters = self.config['include_filters']
        if isinstance(include_filters, basestring):
            include_filters = [include_filters]

        self.include_reg = re.compile(r'(?:%s)' % '|'.join(include_filters))

    def get_default_config_help(self):
        """
        Returns the help text for the configuration options for this handler
        """
        config = super(LibratoHandler, self).get_default_config_help()

        config.update({
            'user': 'Librato username',
            'apikey': 'Librato API key',
            'apply_metric_prefix': 'Allow diamond to apply metric prefix',
            'queue_max_size': 'Max measurements to queue before submitting',
            'queue_max_interval':
                'Max seconds to wait before submitting. For best behavior, '
                'be sure your highest collector poll interval is lower than '
                'or equal to the queue_max_interval setting.',
            'include_filters':
                'A list of regex patterns. Only measurements whose path '
                'matches a filter will be submitted. Useful for limiting '
                'usage to *only* desired measurements, e.g. '
                '`"^diskspace\..*\.byte_avail$", "^loadavg\.01"` or '
                '`"^sockets\.",` (note trailing comma to indicate a list)',
        })

        return config

    def get_default_config(self):
        """
        Return the default config for the handler
        """
        config = super(LibratoHandler, self).get_default_config()

        config.update({
            'user': '',
            'apikey': '',
            'apply_metric_prefix': False,
            'queue_max_size': 300,
            'queue_max_interval': 60,
            'include_filters': ['^.*'],
        })

        return config

    def process(self, metric):
        """
        Process a metric by sending it to Librato
        """
        path = metric.getCollectorPath()
        path += '.'
        path += metric.getMetricPath()
        if self.config['apply_metric_prefix']:
            path = metric.getPathPrefix() + '.' + path

        if self.include_reg.match(path):
            if metric.metric_type == 'GAUGE':
                m_type = 'gauge'
            else:
                m_type = 'counter'
            self.queue.add(path,                # name
                           float(metric.value),  # value
                           type=m_type,
                           source=metric.host,
                           measure_time=metric.timestamp)
            self.current_n_measurements += 1
        else:
            self.log.debug("LibratoHandler: Skip %s, no include_filters match",
                           path)

        if (self.current_n_measurements >= self.queue_max_size or
                time.time() >= self.queue_max_timestamp):
            self.log.debug("LibratoHandler: Sending batch size: %d",
                           self.current_n_measurements)
            self._send()

    def flush(self):
        """Flush metrics in queue"""
        self._send()

    def _send(self):
        """
        Send data to Librato.
        """
        self.queue.submit()
        self.queue_max_timestamp = int(time.time() + self.queue_max_interval)
        self.current_n_measurements = 0



# coding=utf-8

"""
Uses /proc/net/sockstat to collect data on number of open sockets

#### Dependencies

 * /proc/net/sockstat

"""

import diamond.collector
import re
import os
from collections import defaultdict


_RE = re.compile('|'.join([
    r'sockets: used (?P<used>\d+)?',
    r'(TCP|TCP6): inuse (?P<tcp_inuse>\d+)' +
    r'( orphan (?P<tcp_orphan>\d+) ' +
    r'tw (?P<tcp_tw>\d+) ' +
    r'alloc (?P<tcp_alloc>\d+) ' +
    r'mem (?P<tcp_mem>\d+))?',
    r'(UDP|UDP6): inuse (?P<udp_inuse>\d+)( mem (?P<udp_mem>\d+))?'
]))


class SockstatCollector(diamond.collector.Collector):

    PROCS = ['/proc/net/sockstat', '/proc/net/sockstat6']

    def get_default_config_help(self):
        config_help = super(SockstatCollector, self).get_default_config_help()
        config_help.update({
        })
        return config_help

    def get_default_config(self):
        """
        Returns the default collector settings
        """
        config = super(SockstatCollector, self).get_default_config()
        config.update({
            'path':     'sockets',
        })
        return config

    def collect(self):

        result = defaultdict(int)
        for path in self.PROCS:
            if not os.access(path, os.R_OK):
                continue

            f = open(path)
            self.collect_stat(result, f)
            f.close()

        for key, value in result.items():
            self.publish(key, value, metric_type='GAUGE')

    def collect_stat(self, data, f):

        for line in f:
            match = _RE.match(line)
            if match:
                for key, value in match.groupdict().items():
                    if value:
                        data[key] += int(value)



# coding=utf-8

"""
Uses /proc to collect available entropy

#### Dependencies

 * /proc/sys/kernel/random/entropy_avail

"""

import diamond.collector
import os


class EntropyStatCollector(diamond.collector.Collector):

    PROC = '/proc/sys/kernel/random/entropy_avail'

    def get_default_config(self):
        """
        Returns the default collector settings
        """
        config = super(EntropyStatCollector, self).get_default_config()
        config.update({
            'path':     'entropy'
        })
        return config

    def collect(self):
        if not os.access(self.PROC, os.R_OK):
            return None

        # open file
        entropy_file = open(self.PROC)

        # read value
        entropy = entropy_file.read().strip()

        # Close file
        entropy_file.close()

        # Publish value
        self.publish_gauge("available", entropy)



# coding=utf-8

"""
Uses the vms suite to query per instance memory metrics, for VMS-enabled
instances

#### Dependencies

 * vms

"""

# Note, the DIAMOUND_USER has to be added to the VMS_GROUP specified
# in /etc/sysconfig/vms. VMS_GROUP may not be defined in which case it
# defaults to the login group of VMS_USER. If you are running diamond
# as root, no worries. In most other cases:
#     usermod -G kvm diamond
# should suffice

import diamond.collector
try:
    import vms
except ImportError:
    vms = None


class VMSDomsCollector(diamond.collector.Collector):
    PLUGIN_STATS = {
        'nominal': ('pages', 4096),
        'current': ('memory.current', 4096),
        'clean': ('memory.clean', 4096),
        'dirty': ('memory.dirty', 4096),
        'limit': ('memory.limit', 4096),
        'target': ('memory.target', 4096),
        'evicted': ('eviction.dropped', 4096),
        'pagedout': ('eviction.pagedout', 4096),
        'pagedin': ('eviction.pagedin', 4096),
    }

    def get_default_config_help(self):
        config_help = super(VMSDomsCollector, self).get_default_config_help()
        config_help.update({
        })
        return config_help

    def get_default_config(self):
        """
        Returns the default collector settings
        """
        config = super(VMSDomsCollector, self).get_default_config()
        config.update({
            'path':     'vms'
        })
        return config

    def collect(self):
        if vms is None:
            self.log.error('Unable to import vms')
            return {}

        vms.virt.init()
        hypervisor = vms.virt.AUTO.Hypervisor()

        # Get list of domains and iterate.
        domains = hypervisor.domain_list()
        vms_domains = []
        count = 0

        # Pre filter VMS domains.
        for d in domains:
            # Skip non-VMS domains.
            if not vms.control.exists(d):
                continue

            # Grab a control connection.
            dom = hypervisor.domain_lookup(d)
            if dom is None:
                continue
            ctrl = dom._wait_for_control(wait=False)
            if ctrl is None:
                continue

            try:
                # Skip ghost domains.
                if ctrl.get('gd.isghost') == '1':
                    continue
            except vms.control.ControlException:
                continue

            vms_domains.append((dom, ctrl))
            count += 1

        # Add the number of domains.
        self.publish('domains', count)

        # For each stat,
        for stat in self.PLUGIN_STATS:
            key = self.PLUGIN_STATS[stat][0]
            scale = self.PLUGIN_STATS[stat][1]
            total = 0

            # For each domain,
            for dom, ctrl in vms_domains:
                try:
                    # Get value and scale.
                    value = long(ctrl.get(key)) * scale
                except vms.control.ControlException:
                    continue

                # Dispatch.
                self.publish(stat, value, instance=dom.name())

                # Add to total.
                total = total + value

            # Dispatch total value.
            self.publish(stat, total)
